@@ -24,6 +24,7 @@
 #include "ephy-embed-container.h"
 #include "ephy-embed-shell.h"
 #include "ephy-lib-type-builtins.h"
+#include "ephy-output-encoding.h"
 #include "ephy-settings.h"
 #include "ephy-web-view.h"
 
@@ -45,7 +46,6 @@ typedef struct {
   WebKitWebView *web_view;
   GCancellable *cancellable;
   guint load_changed_id;
-  char *uri;
 } EphyReaderRequest;
 
 static EphyReaderRequest *
@@ -70,7 +70,6 @@ ephy_reader_request_free (EphyReaderRequest *request)
   if (request->load_changed_id > 0)
     g_signal_handler_disconnect (request->web_view, request->load_changed_id);
 
-  g_clear_pointer (&request->uri, g_free);
   g_object_unref (request->source_handler);
   g_object_unref (request->scheme_request);
   g_clear_object (&request->web_view);
@@ -124,9 +123,9 @@ enum_nick (GType enum_type,
   return nick;
 }
 
-static
-char *readability_get_property_string (WebKitJavascriptResult *js_result,
-                                       char                   *property)
+static char *
+readability_get_property_string (WebKitJavascriptResult *js_result,
+                                 char                   *property)
 {
   JSCValue *jsc_value;
   char *result = NULL;
@@ -158,7 +157,10 @@ readability_js_finish_cb (GObject      *object,
   g_autoptr (WebKitJavascriptResult) js_result = NULL;
   g_autoptr (GError) error = NULL;
   g_autofree gchar *byline = NULL;
+  g_autofree gchar *encoded_byline = NULL;
   g_autofree gchar *content = NULL;
+  g_autofree gchar *encoded_content = NULL;
+  g_autofree gchar *encoded_title = NULL;
   g_autoptr (GString) html = NULL;
   g_autoptr (GBytes) style_css = NULL;
   const gchar *title;
@@ -175,10 +177,15 @@ readability_js_finish_cb (GObject      *object,
 
   byline = readability_get_property_string (js_result, "byline");
   content = readability_get_property_string (js_result, "content");
-
-  html = g_string_new ("");
-  style_css = g_resources_lookup_data ("/org/gnome/epiphany/readability/reader.css", G_RESOURCE_LOOKUP_FLAGS_NONE, NULL);
   title = webkit_web_view_get_title (web_view);
+
+  encoded_byline = byline ? ephy_encode_for_html_entity (byline) : g_strdup ("");
+  encoded_content = ephy_encode_for_html_entity (content);
+  encoded_title = ephy_encode_for_html_entity (title);
+
+  html = g_string_new (NULL);
+  style_css = g_resources_lookup_data ("/org/gnome/epiphany/readability/reader.css", G_RESOURCE_LOOKUP_FLAGS_NONE, NULL);
+
   font_style = enum_nick (EPHY_TYPE_PREFS_READER_FONT_STYLE,
                           g_settings_get_enum (EPHY_SETTINGS_READER,
                                                EPHY_PREFS_READER_FONT_STYLE));
@@ -199,12 +206,12 @@ readability_js_finish_cb (GObject      *object,
                           "</i>"
                           "<hr>",
                           (gchar *)g_bytes_get_data (style_css, NULL),
-                          title,
+                          encoded_title,
                           font_style,
                           color_scheme,
-                          title,
-                          byline != NULL ? byline : "");
-  g_string_append (html, content);
+                          encoded_title,
+                          encoded_byline);
+  g_string_append (html, encoded_content);
   g_string_append (html, "</article>");
 
   finish_uri_scheme_request (request, g_strdup (html->str), NULL);
@@ -252,43 +259,12 @@ ephy_reader_request_begin_get_source_from_uri (EphyReaderRequest *request,
   webkit_web_view_load_uri (request->web_view, uri);
 }
 
-static void
-web_resource_data_cb (WebKitWebResource *resource,
-                      GAsyncResult      *result,
-                      EphyReaderRequest *request)
-{
-  g_autofree guchar *data = NULL;
-  g_autoptr (GError) error = NULL;
-  gsize length;
-
-  data = webkit_web_resource_get_data_finish (resource, result, &length, &error);
-  if (!data || !*data || length == 0) {
-    g_clear_object (&request->web_view);
-    ephy_reader_request_begin_get_source_from_uri (request, request->uri);
-  } else {
-    ephy_reader_request_begin_get_source_from_web_view (request, request->web_view);
-  }
-}
-
-static void
-ephy_reader_request_begin_get_source_from_initiator (EphyReaderRequest *request,
-                                                     WebKitWebView     *web_view)
-{
-  WebKitWebResource *resource = webkit_web_view_get_main_resource (web_view);
-
-  g_assert (resource);
-  webkit_web_resource_get_data (resource,
-                                request->cancellable,
-                                (GAsyncReadyCallback)web_resource_data_cb,
-                                request);
-}
-
 static void
 ephy_reader_request_start (EphyReaderRequest *request)
 {
   g_autoptr (SoupURI) soup_uri = NULL;
-  const char *modified_uri;
   const char *original_uri;
+  WebKitWebView *web_view;
 
   original_uri = webkit_uri_scheme_request_get_uri (request->scheme_request);
   soup_uri = soup_uri_new (original_uri);
@@ -303,18 +279,28 @@ ephy_reader_request_start (EphyReaderRequest *request)
     return;
   }
 
-  /* Extract URI:
-   * ephy-reader:https://example.com/whatever?xyz into https://example.com/whatever?xyz
-   */
-  modified_uri = soup_uri_to_string (soup_uri, TRUE);
-  g_assert (modified_uri);
+  web_view = webkit_uri_scheme_request_get_web_view (request->scheme_request);
+  if (web_view) {
+    gboolean entering_reader_mode;
 
-  request->uri = g_strdup (modified_uri);
-  request->web_view = webkit_uri_scheme_request_get_web_view (request->scheme_request);
-  if (request->web_view && webkit_web_view_get_main_resource (request->web_view))
-    ephy_reader_request_begin_get_source_from_initiator (request, g_object_ref (request->web_view));
-  else
-    ephy_reader_request_begin_get_source_from_uri (request, request->uri);
+    g_object_get (G_OBJECT (web_view), "entering-reader-mode", &entering_reader_mode, NULL);
+    if (!entering_reader_mode)
+      web_view = NULL;
+  }
+
+  if (web_view) {
+    ephy_reader_request_begin_get_source_from_web_view (request, web_view);
+  } else {
+    const char *source_uri;
+
+    /* Extract URI:
+     * ephy-reader:https://example.com/whatever?xyz into https://example.com/whatever?xyz
+     */
+    source_uri = soup_uri_to_string (soup_uri, TRUE);
+    g_assert (source_uri);
+
+    ephy_reader_request_begin_get_source_from_uri (request, source_uri);
+  }
 
   request->source_handler->outstanding_requests =
     g_list_prepend (request->source_handler->outstanding_requests, request);
